SystemOrganization addCategory: #'Seaside-Cache-NG'!

Object subclass: #WAHashTableSizes
	instanceVariableNames: 'candidate goodPrimes primesToAvoid valuesNotToDivide limit'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Seaside-Cache-NG'!

WAHashTableSizes class
	instanceVariableNames: 'sizes'!

!WAHashTableSizes commentStamp: 'pmm 9/10/2013 15:46' prior: 0!
WAHashTableSizes is a helper class, used by hashedCollections to determine sizes for hash tables.

Public protocol is all class-side:

#goodSizeAtLeast: anInteger 
  answers a "good" integer greater than or equal to the given integer.

An integer is not "good" as a hash table size if it is any of:
* Not prime
* Divides 256**k +- a, for small k and a
* Close to a power of two
* Close to dividing the hashMultiply constant

See Andres Valloud's hashing book, and Knuth TAOCP vol. 3.

This class caches a table of selected good primes within the positive SmallInteger range. When this table must be rebuilt, it uses an instance to compute the table. Primes are selected to keep the table fairly small, with approximately five entries per power of two.

The cached table is ordered, and is searched with a binary search to find the closest good size >= the requested size.

Original implementation by Martin McClure!

WAHashTableSizes class
	instanceVariableNames: 'sizes'!

!WAHashTableSizes class methodsFor: 'public' stamp: 'pmm 9/10/2013 15:46'!
atLeast: lowerLimit
	"Answer the next good size >= lowerlimit.
	If lowerLimit is larger than the largest known good prime,
	just make it odd."

	| primes low mid high prime |
	primes := self sizes.
	low := 1.
	high := primes size.
	lowerLimit > (primes at: high)
		ifTrue: [
			^ lowerLimit even
				ifTrue: [ lowerLimit + 1 ]
				ifFalse: [ lowerLimit ] ].
	[ mid := (high - low) // 2 + low.
	prime := primes at: mid.
	prime < lowerLimit
		ifTrue: [ low := mid ]
		ifFalse: [ high := mid ].
	high - low <= 1
		ifTrue: [ ^ primes at: high ] ] repeat! !

!WAHashTableSizes class methodsFor: 'initialize-release' stamp: 'MartinMcClure 1/17/2010 09:51'!
initialize
	"Throw away any previously-cached sizes, then compute and cache the sizes."
	"WAHashTableSizes initialize"
	
	sizes := nil.
	self sizes! !

!WAHashTableSizes class methodsFor: 'private' stamp: 'pmm 9/10/2013 15:46'!
numValuesPerPower
	"Answer the number of values that should be available in the cached table of primes
	for each power-of-two range. A larger number allows closer sizing for pre-sized collections,
	but results in a larger table that takes longer to search."
	
	^ 5 "Chosen so there will be fewer than 128 primes in the sizes table"! !

!WAHashTableSizes class methodsFor: 'private' stamp: 'MartinMcClure 1/17/2010 09:35'!
sizes
	sizes ~~ nil
		ifFalse: [ sizes := self new computeSizes ].
	^ sizes! !

!WAHashTableSizes methodsFor: 'private' stamp: 'MartinMcClure 1/17/2010 09:43'!
candidateIsGoodPrime
	"Answer true if candidate will make a good hash table size.
	Some integers are rejected:
	* Non-primes
	* Primes which are close to dividing 1664525, the hashMultiply constant
	* Primes which divide 256**k +- a, for small k and a
	See Andres Valloud's hashing book, and Knuth TAOCP volume 3."

	candidate isPrime
		ifFalse: [ ^ false ].
	(primesToAvoid includes: candidate)
		ifTrue: [ ^ false ].
	candidate < 256
		ifTrue: [ ^ true ].	"Small primes cannot satisify divisibility constraints"
	^ valuesNotToDivide allSatisfy: [ :dividend | dividend \\ candidate ~~ 0 ]! !

!WAHashTableSizes methodsFor: 'private' stamp: 'pmm 9/10/2013 15:47'!
computeSizes
	"Answer an array of integers that make good hash table sizes.
	In each power of two, there are about five primes to choose from.
	Some primes are rejected:
	* Primes close to a power of two.
	* Primes which divide 256**k +- a, for small k and a
	* Primes which are close to dividing 1664525, the hashMultiply constant
	See Andres Valloud's hashing book, and Knuth TAOCP volume 3."

	| logInterval |
	logInterval := 0.5 / self numValuesPerPower.
	2 + logInterval to: 30 by: 2 * logInterval do: [ :exp | 
		(self goodPrimeForExp: exp)
			ifNotNil: [ :prime | 
				goodPrimes last ~~ prime
					ifTrue: [ goodPrimes add: prime ] ] ].
	^ goodPrimes asArray! !

!WAHashTableSizes methodsFor: 'private' stamp: 'pmm 9/10/2013 15:47'!
firstCandidateForExp: exp
	"Answer the smallest odd integer greater 2**exp."
	
	| n |
	n := (2 raisedTo: exp) rounded.
	^ n odd 
		ifTrue: [ n ]
		ifFalse: [ n + 1 ]! !

!WAHashTableSizes methodsFor: 'private' stamp: 'MartinMcClure 1/17/2010 09:41'!
goodPrimeForExp: exp
	"Answer the next prime integer >= 2**exp that will make a good hash table size,
	Some primes are rejected:
	* Primes close to a power of two.
	* Primes which divide 256**k +- a, for small k and a
	* Primes which are close to dividing 1664525, the hashMultiply constant
	See Andres Valloud's hashing book, and Knuth TAOCP volume 3."

	candidate := self firstCandidateForExp: exp.
	limit := self limitForExp: exp.
	[ self candidateIsGoodPrime ]
		whileFalse:
			[ candidate := candidate + 2.
			candidate > limit
				ifTrue: [ ^ nil ] ].
	^ candidate! !

!WAHashTableSizes methodsFor: 'initialization' stamp: 'pmm 9/10/2013 15:48'!
initialize
	"Can't use any hashed collections, if sizes is being initialized might get infinite recursion"

	goodPrimes := OrderedCollection new.	
	"Must contain a value less than any prime to avoid extra work in binary search"
	goodPrimes add: 0.
	valuesNotToDivide := OrderedCollection new.
	1 to: 8 do: [ :k | 
		| n |
		n := 256 raisedToInteger: k.
		-32 to: 32 do: [ :a | valuesNotToDivide add: n + a ] ].
	primesToAvoid := self primeAlmostFactorsOf: 1 hashMultiply! !

!WAHashTableSizes methodsFor: 'private' stamp: 'pmm 9/10/2013 15:48'!
limitForExp: exp
	"Answer the largest integer that isn't too close to the next higher power of 2 than exp."
	
	| expLimit |
	expLimit := exp ceiling - (0.5 / self numValuesPerPower).
	^ (2 raisedTo: expLimit) rounded.
	! !

!WAHashTableSizes methodsFor: 'private' stamp: 'pmm 9/10/2013 15:48'!
numValuesPerPower
	"Answer the number of values that should be available in the cached table of primes
	for each power-of-two range."
	
	^ self class numValuesPerPower
! !

!WAHashTableSizes methodsFor: 'private' stamp: 'pmm 9/10/2013 15:48'!
primeAlmostFactorsOf: anInteger
	"Answer primes less than anInteger whose remainder when dividing anInteger is small"

	| factors trial |
	factors := OrderedCollection new.
	anInteger even ifTrue: [factors add: 2].
	3 to: anInteger // 2 + 2 by: 2 do: [:i | 
		(i isPrime and: [| remainder | 
						remainder := anInteger \\ i.
						remainder <= 1 or: [remainder = (i - 1)]])
			ifTrue: [factors add: i]].
	^ factors asArray	
! !

WAObject subclass: #WACacheKeyEntryNG
	instanceVariableNames: 'key keyHash value next byAccessListEntry byCreationListEntry'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Seaside-Cache-NG'!

!WACacheKeyEntryNG class methodsFor: 'instance creation' stamp: 'pmm 9/10/2013 09:08'!
key: aKey value: anObject
	^ self basicNew
		initializeWithKey: aKey value: anObject;
		yourself! !

!WACacheKeyEntryNG methodsFor: 'accessing' stamp: 'pmm 9/8/2013 18:33'!
byAccessListEntry
	^ byAccessListEntry! !

!WACacheKeyEntryNG methodsFor: 'accessing' stamp: 'pmm 9/10/2013 09:21'!
byAccessListEntry: aCacheListEntry
	byAccessListEntry := aCacheListEntry! !

!WACacheKeyEntryNG methodsFor: 'accessing' stamp: 'pmm 9/8/2013 18:33'!
byCreationListEntry
	^ byCreationListEntry! !

!WACacheKeyEntryNG methodsFor: 'accessing' stamp: 'pmm 9/10/2013 09:22'!
byCreationListEntry: aCacheListEntry
	byCreationListEntry := aCacheListEntry! !

!WACacheKeyEntryNG methodsFor: 'initialization' stamp: 'pmm 9/10/2013 09:07'!
initializeWithKey: aKey value: anObject
	self initialize.
	key := aKey.
	keyHash := aKey hash.
	value := anObject! !

!WACacheKeyEntryNG methodsFor: 'accessing' stamp: 'pmm 9/8/2013 18:28'!
key
	^ key! !

!WACacheKeyEntryNG methodsFor: 'accessing' stamp: 'pmm 9/9/2013 11:51'!
keyHash
	^ keyHash! !

!WACacheKeyEntryNG methodsFor: 'accessing' stamp: 'pmm 9/9/2013 11:07'!
next
	^ next! !

!WACacheKeyEntryNG methodsFor: 'accessing' stamp: 'pmm 9/9/2013 14:53'!
next: aCacheKeyEntry
	next := aCacheKeyEntry! !

!WACacheKeyEntryNG methodsFor: 'accessing' stamp: 'pmm 9/8/2013 18:28'!
value
	^ value! !

!WACacheKeyEntryNG methodsFor: 'accessing' stamp: 'pmm 9/9/2013 15:22'!
value: anObject
	value := anObject! !

WAObject subclass: #WACacheListEntryNG
	instanceVariableNames: 'previous next time cacheEntry'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Seaside-Cache-NG'!

!WACacheListEntryNG class methodsFor: 'instance creation' stamp: 'pmm 9/10/2013 09:16'!
entry: aCacheEntry
	^ self basicNew
		initializeWithEntry: aCacheEntry;
		yourself! !

!WACacheListEntryNG methodsFor: 'accessing' stamp: 'pmm 9/8/2013 18:27'!
cacheEntry
	^ cacheEntry! !

!WACacheListEntryNG methodsFor: 'operations' stamp: 'pmm 9/9/2013 12:06'!
expunge
	next isNil ifFalse: [
		next previous: previous ].
	previous isNil ifFalse: [
		previous next: next ].
	next := nil.
	previous := nil! !

!WACacheListEntryNG methodsFor: 'operations' stamp: 'pmm 9/9/2013 12:08'!
initializeTime
	time := Time totalSeconds! !

!WACacheListEntryNG methodsFor: 'initialization' stamp: 'pmm 9/10/2013 09:15'!
initializeWithEntry: aCacheEntry
	self initialize.
	cacheEntry := aCacheEntry.
	self initializeTime! !

!WACacheListEntryNG methodsFor: 'accessing' stamp: 'pmm 9/9/2013 09:28'!
next
	^ next! !

!WACacheListEntryNG methodsFor: 'accessing' stamp: 'pmm 9/9/2013 09:28'!
next: aCacheListEntry
	next := aCacheListEntry! !

!WACacheListEntryNG methodsFor: 'accessing' stamp: 'pmm 9/9/2013 09:28'!
previous
	^ previous! !

!WACacheListEntryNG methodsFor: 'accessing' stamp: 'pmm 9/9/2013 09:28'!
previous: aCacheListEntry
	previous := aCacheListEntry! !

!WACacheListEntryNG methodsFor: 'accessing' stamp: 'pmm 9/8/2013 18:17'!
time
	^ time! !

WAObject subclass: #WACacheNG
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Seaside-Cache-NG'!

!WACacheNG commentStamp: 'pmm 9/9/2013 09:13' prior: 0!
WACacheNG is the abstract base class for session and continuation stores in Seaside. The default implementation is WAHashCacheNG.!

!WACacheNG methodsFor: 'accessing' stamp: 'pmm 8/26/2012 12:17'!
at: aKey ifAbsent: aBlock
	self subclassResponsibility! !

!WACacheNG methodsFor: 'putting' stamp: 'pmm 8/26/2012 12:15'!
at: aKey put: anObject
	self subclassResponsibility! !

!WACacheNG methodsFor: 'public' stamp: 'pmm 9/8/2013 17:50'!
clear
	self subclassResponsibility! !

!WACacheNG methodsFor: 'accessing' stamp: 'pmm 8/26/2012 12:19'!
keyAtValue: anObject ifAbsent: aBlock
	self subclassResponsibility! !

!WACacheNG methodsFor: 'accessing' stamp: 'pmm 9/10/2013 12:06'!
keySize
	^ 16! !

!WACacheNG methodsFor: 'enumerating' stamp: 'pmm 8/26/2012 12:04'!
keysAndValuesDo: aTwoArgumentBlock
	self subclassResponsibility! !

!WACacheNG methodsFor: 'copying' stamp: 'pmm 9/9/2013 09:12'!
postCopy
	self subclassResponsibility! !

!WACacheNG methodsFor: 'removing' stamp: 'pmm 8/26/2012 12:16'!
remove: anObject
	self subclassResponsibility! !

!WACacheNG methodsFor: 'putting' stamp: 'pmm 8/26/2012 12:49'!
store: anObject
	"Store anObject in the cache, generating a new random key. Return the key."
	self subclassResponsibility! !

WACacheNG subclass: #WADictionaryCacheNG
	instanceVariableNames: 'byKey byValue'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Seaside-Cache-NG'!

!WADictionaryCacheNG commentStamp: 'pmm 9/10/2013 11:35' prior: 0!
WADictionaryCacheNG is a dictionary based implementation of WACacheNG.!

!WADictionaryCacheNG methodsFor: 'accessing' stamp: 'pmm 9/10/2013 11:36'!
at: aKey ifAbsent: aBlock
	^ byKey at: aKey ifAbsent: aBlock! !

!WADictionaryCacheNG methodsFor: 'putting' stamp: 'pmm 9/10/2013 11:40'!
at: aKey put: anObject
	byKey at: aKey put: anObject.
	byValue at: anObject put: aKey ! !

!WADictionaryCacheNG methodsFor: 'public' stamp: 'pmm 9/10/2013 11:39'!
clear
	byKey removeAll.
	byValue removeAll! !

!WADictionaryCacheNG methodsFor: 'initialization' stamp: 'pmm 9/10/2013 11:36'!
initialize
	self initialize.
	byKey := Dictionary new.
	byValue := Dictionary new! !

!WADictionaryCacheNG methodsFor: 'accessing' stamp: 'pmm 9/10/2013 11:37'!
keyAtValue: anObject ifAbsent: aBlock
	^ byValue at: anObject ifAbsent: aBlock! !

!WADictionaryCacheNG methodsFor: 'enumerating' stamp: 'pmm 9/10/2013 11:38'!
keysAndValuesDo: aTwoArgumentBlock
	byKey keysAndValuesDo: aTwoArgumentBlock! !

!WADictionaryCacheNG methodsFor: 'copying' stamp: 'pmm 9/10/2013 11:38'!
postCopy
	super postCopy.
	byKey := byKey copy.
	byValue := byValue copy! !

!WADictionaryCacheNG methodsFor: 'removing' stamp: 'pmm 9/10/2013 11:42'!
remove: anObject
	| key |
	key := byValue removeKey: anObject.
	byKey removeKey: key! !

!WADictionaryCacheNG methodsFor: 'putting' stamp: 'pmm 9/10/2013 11:41'!
store: anObject
	| candidateKey |
	[ 
		candidateKey := WAKeyGenerator current keyOfLength: self keySize.
		self includesKey: candidateKey ] whileTrue.
	self at: candidateKey put: anObject.
	^ candidateKey! !

WACacheNG subclass: #WAHashCacheNG
	instanceVariableNames: 'keyTable valueTable size maximumSize maximumRelativeAge maximumAbsoluteAge overflowAction byAccessStart byAccessEnd byCreationStart byCreationEnd'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Seaside-Cache-NG'!

!WAHashCacheNG commentStamp: 'pmm 9/9/2013 12:01' prior: 0!
WAHashCacheNG is a hashtable based implementation of WACacheNG.

The characteristics of WAHashCacheNG are:
- supports both absolute and relative timeouts at the same time https://www.owasp.org/index.php/Session_Management_Cheat_Sheet#Automatic_Session_Expiration
- supports a maximum size with definable overflow action (expire oldest, expire least recently used, signal exception)
- access by key is fast (O(1) average case O(n) worst case)
- access by value is fast (O(1) average case O(n) worst case)
- reaping expired sesssions is proportinal (O(n)) to the number of expired session and independent of the total number of sessions (O(1))
- creating a new session independent of the total number of sessions

Instance Variables:
	keyTable	<Matrix | SequenceableCollection>
	valueTable	<Matrix | SequenceableCollection>
	size	<Color | DateAndTime | Duration | FloatArray | Interval | Number | OBMetaNode | Point | Timespan | TraitComposition | TraitDescription | TraitTransformation | WordArray>
	maximumSize	<ProtoObject | PseudoContext>
	maximumRelativeAge	<CharacterBlock | GRVersion | Magnitude | Point | String | UUID | WAHttpVersion>
	maximumAbsoluteAge	<CharacterBlock | GRVersion | Magnitude | Point | String | UUID | WAHttpVersion>
	overflowAction	<ProtoObject | PseudoContext>
	byAccessRoot	<ProtoObject | PseudoContext>
	byCreationRoot	<ProtoObject | PseudoContext>!

!WAHashCacheNG class methodsFor: 'instance creation' stamp: 'pmm 9/10/2013 11:34'!
initialSize: anSizeInteger maximumSize: aMaximumSizeInteger maximumAbsoluteAge: aMaximumAbsoluteAge maximumRelativeAge: aMaximumRelativeAge overflowAction: aOverflowActionSymbol
	^ self basicNew
		initializeWithInitialSize: anSizeInteger maximumSize: aMaximumSizeInteger maximumAbsoluteAge: aMaximumAbsoluteAge maximumRelativeAge: aMaximumRelativeAge overflowAction: aOverflowActionSymbol;
		yourself! !

!WAHashCacheNG class methodsFor: 'accessing' stamp: 'pmm 9/8/2013 18:16'!
overflowActions
	^ Array
		with: self removeRelativeOldest
		with: self removeAbsoluteOldest
		with: self signalError! !

!WAHashCacheNG class methodsFor: 'accessing' stamp: 'pmm 9/8/2013 18:15'!
removeAbsoluteOldest
	^ #removeAbsoluteOldest! !

!WAHashCacheNG class methodsFor: 'accessing' stamp: 'pmm 9/8/2013 18:15'!
removeRelativeOldest
	^ #removeRelativeOldest! !

!WAHashCacheNG class methodsFor: 'accessing' stamp: 'pmm 9/8/2013 18:01'!
signalError
	^ #signalError! !

!WAHashCacheNG methodsFor: 'private' stamp: 'pmm 9/9/2013 15:33'!
addToAccessEnd: aCacheListEntry
	| currentLast |
	currentLast := byAccessEnd.
	currentLast isNil ifFalse: [
		currentLast next: aCacheListEntry.
		aCacheListEntry previous: currentLast ].
	aCacheListEntry initializeTime.
	byAccessEnd := aCacheListEntry! !

!WAHashCacheNG methodsFor: 'private' stamp: 'pmm 9/9/2013 15:32'!
addToCreationEnd: aCacheListEntry
	| currentLast |
	currentLast := byCreationEnd.
	currentLast isNil ifFalse: [
		currentLast next: aCacheListEntry.
		aCacheListEntry previous: currentLast ].
	byCreationEnd := aCacheListEntry! !

!WAHashCacheNG methodsFor: 'accessing' stamp: 'pmm 9/10/2013 12:33'!
at: aKey ifAbsent: aBlock
	| keyHash index entry |
	self sweepIncrementally.
	keyHash := aKey hash.
	index := (keyHash \\ keyTable size) + 1.
	entry := self findEntryOrNilWithKey: aKey hash: keyHash index: index.
	^ entry isNil
		ifTrue: [ aBlock value ]
		ifFalse: [ entry value ]! !

!WAHashCacheNG methodsFor: 'putting' stamp: 'pmm 9/10/2013 14:01'!
at: aKey put: anObject
	| index keyHash entry |
	self sweepIncrementally.
	self growCheck.
	keyHash := aKey hash.
	index := (keyHash \\ keyTable size) + 1.
	entry := self findEntryOrNilWithKey: aKey hash: keyHash index: index.
	entry isNil
		ifTrue: [
			| cacheEntry byAccessEntry byCreationEntry |
			(maximumSize > 0 and: [ size = maximumSize ])
				ifTrue: [ self performOverflowAction ].
			"create entry"
			cacheEntry := WACacheKeyEntryNG key: aKey value: anObject.
			byCreationEntry := WACacheListEntryNG entry: cacheEntry.
			byAccessEntry := WACacheListEntryNG entry: cacheEntry.
			cacheEntry byAccessListEntry: byAccessEntry.
			cacheEntry byCreationListEntry: byCreationEntry.
			self addToAccessEnd: byAccessEntry.
			self addToCreationEnd: byCreationEntry.
			self insertKeyEntry: cacheEntry.
			"update value table"
			self mapValue: anObject to: aKey expectMapping: false.
			size := size + 1 ]
		ifFalse: [
			| byAccessListEntry byCreationListEntry |
			byAccessListEntry := entry byAccessListEntry.
			byCreationListEntry := entry byCreationListEntry.
			self expunge: byAccessListEntry.
			self addToAccessEnd: byAccessListEntry.
			self expunge: byCreationListEntry.
			self addToCreationEnd: byCreationListEntry.
			self updateMappingOf: aKey from: entry value to: anObject.
			
			"update entry"
			entry value: anObject ].
		
	^ anObject! !

!WAHashCacheNG methodsFor: 'public' stamp: 'pmm 9/9/2013 12:01'!
clear
	size = 0.
	keyTable atAllPut: nil.
	valueTable atAllPut: nil. 
	byAccessStart := nil.
	byAccessEnd := nil.
	byCreationStart := nil.
	byCreationEnd := nil! !

!WAHashCacheNG methodsFor: 'private' stamp: 'pmm 9/10/2013 11:45'!
errorAlreadyMapped
	self error: 'Alread mapped'! !

!WAHashCacheNG methodsFor: 'private' stamp: 'pmm 9/9/2013 15:14'!
errorKeyNotFound
	self error: 'Key not found'! !

!WAHashCacheNG methodsFor: 'private' stamp: 'pmm 9/10/2013 11:53'!
errorNotMapped
	self error: 'Not mapped'! !

!WAHashCacheNG methodsFor: 'private' stamp: 'pmm 9/9/2013 12:03'!
expunge: aListEntry
	"update list start"
	aListEntry == byAccessStart
		ifTrue: [ byAccessStart := aListEntry next ].
	aListEntry == byCreationStart
		ifTrue: [ byCreationStart := aListEntry next ].
	"update list end"
	aListEntry == byAccessEnd
		ifTrue: [ byAccessEnd := aListEntry previous ].
	aListEntry == byCreationEnd
		ifTrue: [ byCreationEnd := aListEntry previous ].
	"fix previous and next pointers"
	aListEntry expunge.! !

!WAHashCacheNG methodsFor: 'private' stamp: 'pmm 9/10/2013 12:20'!
findEntryOrNilWithKey: aKey hash: anHashInteger index: anIndexInteger
	| entry |
	entry := keyTable at: anIndexInteger.
	[ entry isNil ] whileFalse: [
		(entry keyHash = anHashInteger and: [ entry key = aKey ])
			ifTrue: [
				| byAccessListEntry |
				byAccessListEntry := entry byAccessListEntry.
				self expunge: byAccessListEntry.
				self addToAccessEnd: byAccessListEntry.
				^ entry ]
			ifFalse: [ entry := entry next ] ].
	^ nil! !

!WAHashCacheNG methodsFor: 'private' stamp: 'pmm 9/10/2013 15:49'!
growCheck
	size > (keyTable size * self loadFactor) ifTrue: [
		"rehash for fun"
		| newKeyTable newValueTable newTableSize |
		newTableSize := WAHashTableSizes atLeast: keyTable size * 2.
		newKeyTable := Array new: newTableSize.
		newValueTable := Array new: newTableSize.
		
		"refash keys"
		keyTable do: [ :each |
			| next current |
			current := each.
			[ current isNil ] whileFalse: [
				| index atTargetIndex |
				next := current next.
				index := (current keyHash \\ newTableSize) + 1.
				atTargetIndex isNil
					ifTrue: [ newKeyTable at: index put: current ]
					ifFalse: [
						[ atTargetIndex next isNil ] whileFalse: [
							atTargetIndex := atTargetIndex next ].
						atTargetIndex next: current ].
				current next: nil.
				current := next ] ].
		
		"rehash values"
		valueTable do: [ :each |
			| next current |
			current := each.
			[ current isNil ] whileFalse: [
				| index atTargetIndex |
				next := current next.
				index := (current valueHash \\ newTableSize) + 1.
				atTargetIndex isNil
					ifTrue: [ newValueTable at: index put: current ]
					ifFalse: [
						[ atTargetIndex next isNil ] whileFalse: [
							atTargetIndex := atTargetIndex next ].
						atTargetIndex next: current ].
				current next: nil.
				current := next ] ].
		keyTable := newKeyTable.
		valueTable := newValueTable ]! !

!WAHashCacheNG methodsFor: 'private' stamp: 'pmm 9/9/2013 14:27'!
includesKey: aKey
	self at: aKey ifAbsent: [ ^ false ].
	^ true! !

!WAHashCacheNG methodsFor: 'initialization' stamp: 'pmm 9/10/2013 11:33'!
initializeWithInitialSize: anSizeInteger maximumSize: aMaximumSizeInteger maximumAbsoluteAge: aMaximumAbsoluteAge maximumRelativeAge: aMaximumRelativeAge overflowAction: aOverflowActionSymbol
	self initialize.
	size := 0.
	keyTable := Array new: anSizeInteger.
	valueTable := Array new: anSizeInteger.
	maximumSize := aMaximumSizeInteger.
	maximumAbsoluteAge := aMaximumAbsoluteAge.
	maximumRelativeAge := aMaximumRelativeAge.! !

!WAHashCacheNG methodsFor: 'private' stamp: 'pmm 9/10/2013 12:58'!
insertKeyEntry: aCacheEntry
	| keyHash key index entry previous |
	keyHash := aCacheEntry keyHash.
	key := aCacheEntry key.
	index := (keyHash \\ valueTable size) + 1.
	entry := keyTable at: index.
	previous := nil.
	[ entry isNil ] whileFalse: [
		(entry keyHash = aCacheEntry and: [ entry key = key ])
			ifTrue: [
				previous isNil
					ifTrue: [ keyTable at: index put: aCacheEntry ]
					ifFalse: [  previous next: aCacheEntry ].
				^ self ]
			ifFalse: [
				previous := entry.
				entry := entry next ] ].
	previous isNil
		ifTrue: [ keyTable at: index put: aCacheEntry ]
		ifFalse: [  previous next: aCacheEntry ]! !

!WAHashCacheNG methodsFor: 'accessing' stamp: 'pmm 9/10/2013 12:39'!
keyAtValue: anObject ifAbsent: aBlock
	| valueHash index entry |
	self sweepIncrementally.
	valueHash := anObject hash.
	index := (valueHash \\ valueTable size) + 1.
	entry := valueTable at: index.
	[ entry isNil ] whileFalse: [
		(entry valueHash = valueHash and: [ entry value = anObject ])
			ifTrue: [
				"update access time"
				self at: entry key ifAbsent: [ self errorKeyNotFound ].
				^ entry key ]
			ifFalse: [ entry := entry next ] ].
	^ aBlock value! !

!WAHashCacheNG methodsFor: 'enumerating' stamp: 'pmm 9/10/2013 12:21'!
keysAndValuesDo: aTwoArgumentBlock
	self sweepIncrementally.
	keyTable do: [ :each |
		| cacheEntry |
		cacheEntry := each.
		[ cacheEntry isNil ] whileFalse: [
			aTwoArgumentBlock value: cacheEntry key value: cacheEntry value.
			cacheEntry := cacheEntry next ] ]! !

!WAHashCacheNG methodsFor: 'private' stamp: 'pmm 9/9/2013 09:17'!
loadFactor
	^ 0.75! !

!WAHashCacheNG methodsFor: 'private' stamp: 'pmm 9/10/2013 12:35'!
mapValue: anObject to: aKey expectMapping: aBoolean
	| valueHash index entry previous newEntry |
	valueHash := anObject hash.
	index := (valueHash \\ valueTable size) + 1.
	newEntry := WACacheValueEntryNG value: anObject key: aKey.
	entry := valueTable at: index.
	previous := nil.
	[ entry isNil ] whileFalse: [
		(entry valueHash = valueHash and: [ entry value = anObject ])
			ifTrue: [
				aBoolean ifFalse: [ self errorAlreadyMapped ].
				previous isNil
					ifTrue: [ valueTable at: index put: newEntry ]
					ifFalse: [  previous next: newEntry ].
				^ self ]
			ifFalse: [
				previous := entry.
				entry := entry next ] ].
	
	aBoolean ifTrue: [ self errorNotMapped ].
	previous isNil
			ifTrue: [ valueTable at: index put: newEntry ]
			ifFalse: [  previous next: newEntry ]! !

!WAHashCacheNG methodsFor: 'private' stamp: 'pmm 9/9/2013 12:01'!
performOverflowAction
	overflowAction == self class signalError ifTrue: [
		WAMaximumNumberOfSessionsExceededError signal ].
	overflowAction == self class removeAbsoluteOldest ifTrue: [
		self removeCacheEntry: byCreationStart cacheEntry ].
	overflowAction == self class removeRelativeOldest ifTrue: [
		self removeCacheEntry: byAccessStart cacheEntry ]! !

!WAHashCacheNG methodsFor: 'copying' stamp: 'pmm 9/10/2013 11:37'!
postCopy
	self notYetImplemented! !

!WAHashCacheNG methodsFor: 'removing' stamp: 'pmm 9/9/2013 11:14'!
remove: anObject
	| key |
	self sweepIncrementally.
	key := self keyAtValue: anObject ifAbsent: [ ^ anObject ].
	self removeKey: key! !

!WAHashCacheNG methodsFor: 'private' stamp: 'pmm 9/9/2013 14:46'!
removeCacheEntry: aCacheEntry
	self removeFromKeyTable: aCacheEntry.
	self removeFromValueTable: aCacheEntry value.
	self expunge: aCacheEntry byAccessListEntry.
	self expunge: aCacheEntry byCreationListEntry.
	
	size := size - 1! !

!WAHashCacheNG methodsFor: 'private' stamp: 'pmm 9/10/2013 12:56'!
removeFromKeyTable: aCacheEntry
	| index entry previous |
	index := (aCacheEntry keyHash \\ keyTable size) + 1.
	entry := keyTable at: index.
	previous := nil.
	"no need for nil check, it has to be here"
	[ aCacheEntry == entry ] whileFalse: [
		previous := entry.
		entry := entry next ].
	previous isNil
		ifTrue: [ keyTable at: index put: entry next ]
		ifFalse: [ previous next: entry next ]! !

!WAHashCacheNG methodsFor: 'private' stamp: 'pmm 9/10/2013 12:35'!
removeFromValueTable: anObject
	| entry valueHash index previous |
	valueHash := anObject hash.
	index := (valueHash \\ valueTable size) + 1.
	entry := valueTable at: index.
	previous := nil.
	[	"no need for nil check it has to be there"
		(entry valueHash = valueHash and: [ entry value = anObject ])
			ifTrue: [
				 previous isNil
					ifTrue: [ valueTable at: index put: entry next ]
					ifFalse: [ previous next: entry next ].
				^ self ]
			ifFalse: [
				previous := entry.
				entry := entry next ] ] repeat.! !

!WAHashCacheNG methodsFor: 'removing' stamp: 'pmm 9/10/2013 12:35'!
removeKey: aKey
	| keyHash index entry |
	self sweepIncrementally.
	keyHash := aKey hash.
	index := (keyHash \\ keyTable size) + 1.
	entry := self findEntryOrNilWithKey: aKey hash: keyHash index: index.
	^ entry isNil
		ifTrue: [ self errorKeyNotFound ]
		ifFalse: [ self removeCacheEntry: entry ]! !

!WAHashCacheNG methodsFor: 'putting' stamp: 'pmm 8/26/2012 12:53'!
store: anObject
	| candidateKey |
	[ 
		candidateKey := WAKeyGenerator current keyOfLength: self keySize.
		self includesKey: candidateKey ] whileTrue.
	self at: candidateKey put: anObject.
	^ candidateKey! !

!WAHashCacheNG methodsFor: 'private' stamp: 'pmm 9/10/2013 12:20'!
sweep: aListRoot maximumAge: anNumberOfSeconds now: currentSeconds
	| current |
	current := aListRoot.
	[ current isNil ] whileFalse: [
		(current time + anNumberOfSeconds) > currentSeconds
			ifTrue: [
				self removeCacheEntry: current cacheEntry ]
			ifFalse: [
				"the list is ordered so we can stop scanning"
				^ self ] ]! !

!WAHashCacheNG methodsFor: 'private' stamp: 'pmm 9/9/2013 12:01'!
sweepIncrementally
	| now |
	now := Time totalSeconds.
	
	"absolute timeout sweep"
	maximumAbsoluteAge > 0 ifTrue: [
		self sweep: byCreationStart maximumAge: maximumAbsoluteAge now: now ].
	
	"relative timeout sweep"
	maximumRelativeAge > 0 ifTrue: [
		self sweep: byAccessStart maximumAge: maximumRelativeAge now: now ]! !

!WAHashCacheNG methodsFor: 'private' stamp: 'pmm 9/10/2013 12:46'!
updateMappingOf: aKey from: oldValue to: newValue
	self removeFromValueTable: oldValue.
	self mapValue: newValue to: aKey expectMapping: false! !

WACacheNG subclass: #WAMutualExclusionCacheNG
	instanceVariableNames: 'mutex cache'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Seaside-Cache-NG'!

!WAMutualExclusionCacheNG commentStamp: 'pmm 9/9/2013 09:07' prior: 0!
WAMutualExclusionCacheNG is a wrapper around a WACacheNG that wraps all messages in a mutex.

Instance Variables:
	mutex	<WAMutex>
	cache	<WACache>!

!WAMutualExclusionCacheNG class methodsFor: 'instance creation' stamp: 'pmm 8/26/2012 12:08'!
on: aCache
	^ self basicNew initializeOn: aCache; yourself! !

!WAMutualExclusionCacheNG methodsFor: 'accessing' stamp: 'pmm 9/8/2013 17:48'!
at: aKey ifAbsent: aBlock
	^ mutex critical: [
		cache at: aKey ifAbsent: aBlock ]	! !

!WAMutualExclusionCacheNG methodsFor: 'putting' stamp: 'pmm 9/8/2013 17:49'!
at: aKey put: anObject
	^ mutex critical: [
		cache at: aKey put: anObject ]	! !

!WAMutualExclusionCacheNG methodsFor: 'public' stamp: 'pmm 9/8/2013 17:49'!
clear
	^ mutex critical: [
		cache clear ]	! !

!WAMutualExclusionCacheNG methodsFor: 'initialization' stamp: 'pmm 9/1/2012 11:07'!
initializeOn: aCache
	self initialize.
	mutex := WAMutex new.
	cache := aCache! !

!WAMutualExclusionCacheNG methodsFor: 'accessing' stamp: 'pmm 9/8/2013 17:48'!
keyAtValue: anObject ifAbsent: aBlock
	^ mutex critical: [
		cache keyAtValue: anObject ifAbsent: aBlock ]	! !

!WAMutualExclusionCacheNG methodsFor: 'accessing' stamp: 'pmm 9/10/2013 12:07'!
keySize
	^ mutex critical: [
		cache keySize ]	! !

!WAMutualExclusionCacheNG methodsFor: 'enumerating' stamp: 'pmm 9/8/2013 17:48'!
keysAndValuesDo: aTwoArgumentBlock
	^ mutex critical: [
		cache keysAndValuesDo: aTwoArgumentBlock ]	! !

!WAMutualExclusionCacheNG methodsFor: 'removing' stamp: 'pmm 9/8/2013 17:50'!
remove: anObject
	^ mutex critical: [
		cache remove: anObject ]	! !

!WAMutualExclusionCacheNG methodsFor: 'putting' stamp: 'pmm 8/26/2012 12:48'!
store: anObject
	^ mutex critical: [
		cache store: anObject ]	! !

WAObject subclass: #WACacheValueEntryNG
	instanceVariableNames: 'value valueHash key next'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Seaside-Cache-NG'!

!WACacheValueEntryNG class methodsFor: 'instance creation' stamp: 'pmm 9/10/2013 09:12'!
value: anObject key: aKey
	^ self basicNew
		initializeWithValue: anObject key: aKey
		yourself! !

!WACacheValueEntryNG methodsFor: 'initialization' stamp: 'pmm 9/10/2013 09:12'!
initializeWithValue: anObject key: aKey
	self initialize.
	value := anObject.
	valueHash := anObject hash.
	key := aKey! !

!WACacheValueEntryNG methodsFor: 'accessing' stamp: 'pmm 9/9/2013 13:26'!
key
	^ key! !

!WACacheValueEntryNG methodsFor: 'accessing' stamp: 'pmm 9/9/2013 13:26'!
next
	^ next! !

!WACacheValueEntryNG methodsFor: 'accessing' stamp: 'pmm 9/9/2013 15:08'!
next: aCacheValueEntry
	next := aCacheValueEntry! !

!WACacheValueEntryNG methodsFor: 'accessing' stamp: 'pmm 9/9/2013 13:27'!
value
	^ value! !

!WACacheValueEntryNG methodsFor: 'accessing' stamp: 'pmm 9/9/2013 13:26'!
valueHash
	^ valueHash! !

WAError subclass: #WAMaximumNumberOfSessionsExceededError
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Seaside-Cache-NG'!

!WAMaximumNumberOfSessionsExceededError commentStamp: 'pmm 9/8/2013 17:59' prior: 0!
http://ora-00018.ora-code.com/!
WAHashTableSizes initialize!
